import threading
import time

import requests

from .config import HUB_API_ROOT
from .utils.general import PREFIX


class HUBLogger:
    api_url = HUB_API_ROOT + "model-checkpoint"

    def __init__(self, model_id, auth):
        self.model_id = model_id
        self.auth = auth
        self.rate_limit = 60.0  # minimum seconds between checkpoint uploads
        self.t = time.time()  # last upload time
        self.keys = [
            'train/box_loss',
            'train/obj_loss',
            'train/cls_loss',  # train loss
            'metrics/precision',
            'metrics/recall',
            'metrics/mAP_0.5',
            'metrics/mAP_0.5:0.95',  # metrics
            'val/box_loss',
            'val/obj_loss',
            'val/cls_loss',  # val loss
            'x/lr0',
            'x/lr1',
            'x/lr2']  # metrics keys

    def on_fit_epoch_end(self, *args, **kwargs):
        """
        Runs after each val each epoch

        Args:
             *args     log_vals, epoch
        """
        vals, epoch = args[:2]
        metrics = {k: float(v) for k, v in zip(self.keys, vals)}  # dict
        # print(f"{PREFIX}Uploading metrics for {self.model_id}: {metrics}")  # debug
        threading.Thread(target=self._upload_epoch, args=(epoch, metrics), daemon=True).start()

    def _upload_epoch(self, epoch, metrics):
        payload = {"modelId": self.model_id, "epoch": epoch, "metrics": metrics, "type": "metrics"}
        payload.update(self.auth.get_auth_string())
        requests.post(self.api_url, data=payload, files={"void": None})

    def on_model_save(self, *args, **kwargs):
        """
        Runs after each model save

        Args:
             *args     last, epoch, final_epoch, best_fitness, fi
        """

        # Set args
        last, epoch, final_epoch, best_fitness, fi = args[:5]
        is_best = best_fitness == fi

        if (time.time() - self.t) > self.rate_limit:
            print(f"{PREFIX}Uploading checkpoint for {self.model_id}")
            threading.Thread(target=self._upload_model, args=(epoch, is_best, last), daemon=True).start()
            self.t = time.time()

    def _upload_model(self, epoch, is_best, weights):
        payload = {"modelId": self.model_id, "epoch": epoch, "isBest": bool(is_best), "type": "epoch"}
        payload.update(self.auth.get_auth_string())
        with open(weights, "rb") as last:
            r = requests.post(self.api_url, data=payload, files={"last.pt": last})
            if r.status_code != 200:
                print(f"{PREFIX}Unable to upload checkpoint!")

    def on_pretrain_routine_end(self, *args, **kwargs):
        # Tell HUB that training has started
        payload = {"modelId": self.model_id, "type": "initial"}
        payload.update(self.auth.get_auth_string())
        requests.post(self.api_url, data=payload, files={"void": None})

    def on_train_end(self, *args, **kwargs):
        """"
        Args:

            *args   last, best, plots, epoch
        """
        # Set args
        last, best, plots, epoch, results = args[:5]

        # mAP0.5:0.95
        map_new = results[3]

        payload = {"modelId": self.model_id, "epoch": epoch, "map": map_new, "type": "final"}
        payload.update(self.auth.get_auth_string())

        print(f"{PREFIX}Uploading final models for {self.model_id}")
        with open(best, "rb") as best:
            r = requests.post(self.api_url, data=payload, files={"best.pt": best})
            if r.status_code != 200:
                print(f"{PREFIX}Unable to upload final models!")



import getpass
import json

import requests

from .config import HUB_API_ROOT
from .utils.general import PREFIX, emojis


class Auth:
    secure_token = id_token = api_key = False

    def __init__(self):
        return

    def attempt_signin(self, attempts=1):
        if attempts > 1:
            tries = "Attempt " + str(attempts) + " of 3"
        else:
            tries = ""
        print(f"{PREFIX}Login. " + tries)
        email = input(f"{PREFIX}Enter your account email address: ")
        password = getpass.getpass("Enter your account email password: ")
        self.signin_email(email, password)
        if not self.id_token:
            attempts += 1

            print(f"{PREFIX}Incorrect Login Details.\n")

            if attempts <= 3:
                return self.attempt_signin(attempts)
            else:
                print(f"{PREFIX}Failed to authenticate. Exiting...")
                return False
        else:
            return True

    def signin_email(self, email, password):
        firebase_rest_url = "https://identitytoolkit.googleapis.com/v1/accounts:signInWithPassword"
        firebase_web_api_key = "AIzaSyDrFrqlV1sr3yK_-w8-cwrKGS9SLdEjW3U"

        payload = json.dumps({
            "email": email,
            "password": password,
            "returnSecureToken": True,})

        r = requests.post(firebase_rest_url, params={"key": firebase_web_api_key}, data=payload)

        if "error" in r.json():
            self.secure_token = False
            self.id_token = False
        else:
            self.secure_token = r.json()
            self.id_token = self.secure_token.get("idToken")

    def attempt_api_key(self, api_key='', attempts=1):
        if not api_key:
            s = f'. Attempt {attempts} of 3' if attempts > 1 else ' from https://hub.ultralytics.com/settings/api-keys'
            print(PREFIX, end='')
            api_key = getpass.getpass(f"Enter API key{s}: ")
        self.validate_api_key(api_key)
        if not self.api_key:
            attempts += 1
            print(emojis(f"{PREFIX}Invalid API key ⚠️\n"))

            if attempts <= 3:
                return self.attempt_api_key(attempts=attempts)
            else:
                print(emojis(f"{PREFIX}Failed to authenticate ❌"))
                return False
        else:
            print(emojis(f"{PREFIX}Authenticated ✅"))
            return True

    def validate_api_key(self, api_key):
        auth_endpoint = HUB_API_ROOT + "/authorise"

        payload = {"apiKey": api_key}

        r = requests.post(auth_endpoint, json=payload)

        self.secure_token = False
        self.id_token = False

        if r.status_code == 200:
            self.api_key = api_key
        else:
            self.api_key = False

    def get_auth_string(self):
        if self.id_token:
            return {"idToken": self.id_token}
        elif self.api_key:
            return {"apiKey": self.api_key}
        else:
            return None



__version__ = '0.0.15'  # print(ultralytics.__version__)

from .utils.general import PREFIX, check_requirements, colorstr, emojis, is_colab


def checks(verbose=True):
    # Check system software and hardware
    from .yolov5_wrapper import clone_yolo

    clone_yolo()

    print('Checking setup...')

    import os
    import shutil

    from .utils.torch_utils import select_device  # imports

    check_requirements(('psutil', 'IPython'))
    import psutil
    from IPython import display  # to display images and clear console output

    if is_colab():
        shutil.rmtree('sample_data', ignore_errors=True)  # remove colab /sample_data directory

    if verbose:
        # System info
        # gb = 1 / 1000 ** 3  # bytes to GB
        gib = 1 / 1024 ** 3  # bytes to GiB
        ram = psutil.virtual_memory().total
        total, used, free = shutil.disk_usage("/")
        display.clear_output()
        s = f'({os.cpu_count()} CPUs, {ram * gib:.1f} GB RAM, {(total - free) * gib:.1f}/{total * gib:.1f} GB disk)'
    else:
        s = ''

    select_device(newline=False, version=__version__)
    print(emojis(f'Setup complete ✅ {s}'))


def login(api_key=''):
    # Login to Ultralytics HUB
    from .main import connect_to_hub
    connect_to_hub(api_key, verbose=True)


def start(key=''):
    # Start training models with Ultralytics HUB
    # Test usage: from src.ultralytics import start; start('API_KEY')
    import getpass
    from .main import train_model

    s = emojis(f"{PREFIX}Invalid API key ⚠️\n")  # error string
    if not key:
        key = getpass.getpass(f"Enter model key: ")
    assert '.' in key, s
    api_key, model_key = key.split('.')
    assert len(api_key) and len(model_key), s

    # key = 'ac0ab020186aeb50cc4c2a5272de17f58bbd2c0.RqFCDNBxgU4mOLmaBrcd'  # example
    # api_key='ac0ab020186aeb50cc4c2a5272de17f58bbd2c0', model_key='RqFCDNBxgU4mOLmaBrcd'  # example
    train_model(api_key=api_key, model_key=model_key)



# Trains and Updates Ultralytics HUB Models

from .auth import Auth
from .trainer import Trainer
from .utils.general import PREFIX
from .yolov5_wrapper import clone_yolo

AUTH = Auth()
CONNECTED = False


def train_model(api_key='', model_key='') -> None:
    """Starts training from next in queue"""
    clone_yolo()
    # connect_to_hub(api_key=api_key)  # deprecated
    AUTH.api_key = api_key
    trainer = Trainer(model_id=model_key, auth=AUTH)  # No model so next in train queue is fetched
    if trainer.model is not None:
        trainer.start()


def connect_to_hub(api_key='', verbose=False) -> bool:
    """Authenticates user with Ultralytics HUB"""
    global CONNECTED

    if CONNECTED and verbose:
        print(f'{PREFIX}Already logged in.')
    elif not CONNECTED:
        CONNECTED = AUTH.attempt_api_key(api_key=api_key)

    return CONNECTED



import importlib  # To import YOLOv5 as module
import subprocess
from pathlib import Path

from .config import ENVIRONMENT, REPO_BRANCH, REPO_URL
from .utils.general import emojis


def install_dependencies():
    print(emojis('Installing YOLOv5 🚀 requirements...'), end=" ")
    s = ['pip', 'install', '-r', 'yolov5/requirements.txt']
    if ENVIRONMENT == 'development':
        subprocess.call(s)
    else:
        subprocess.call(s, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    print('Done')


def clone_yolo():
    """
    Check that YOLOv5 is on the system and clone if not found
    """
    if not Path("./yolov5").is_dir():
        from git import Repo
        try:
            print(emojis(f"Cloning YOLOv5 🚀 from {REPO_URL}..."), end=" ")
            Repo.clone_from(REPO_URL, "yolov5", branch=REPO_BRANCH)  # Temp solution while waiting on pull request
            print("Done")
            install_dependencies()

        except Exception as e:
            print(f"ERROR: Failed to clone YOLOv5: {e}")


class YOLOv5Wrapper:
    @staticmethod
    def export(**kwargs):
        """Calls the YOLOv5 export module"""
        _export = importlib.import_module("yolov5.export")
        _export.run(**kwargs)

    @staticmethod
    def detect(**kwargs):
        """Calls the YOLOv5 detect module"""
        _detect = importlib.import_module("yolov5.detect")
        _detect.run(**kwargs)

    @staticmethod
    def train(callback_handler, **kwargs):
        """Calls the YOLOv5 train module"""
        _train = importlib.import_module("yolov5.train")
        opt = _train.parse_opt(True)
        for k, v in kwargs.items():
            setattr(opt, k, v)

        # _train.run(**kwargs)
        # text_trap = io.StringIO()
        # sys.stdout = text_trap # Trap output
        _train.main(opt, callback_handler)
        # sys.stdout = sys.__stdout__ # Restore output

    @staticmethod
    def new_callback_handler():
        """Returns a YOLOv5 callback handler"""
        # _callback_handler = importlib.import_module("yolov5.utils.callbacks")
        # return _callback_handler.Callbacks()  # DOUBLE-CHECKPOINT FIX
        from yolov5.utils.callbacks import Callbacks  # assume yolov5/ in cwd
        return Callbacks()



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
General utils
"""

import logging
import os
import platform
from pathlib import Path
from subprocess import check_output

import pkg_resources as pkg

# Settings
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # YOLOv5 root directory
VERBOSE = str(os.getenv('YOLOv5_VERBOSE', True)).lower() == 'true'  # global verbose mode


def colorstr(*input):
    # Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])  # color arguments, string
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']


def emojis(str=''):
    # Return platform-dependent emoji-safe version of string
    return str.encode().decode('ascii', 'ignore') if platform.system() == 'Windows' else str


def is_colab():
    # Is environment a Google Colab instance?
    try:
        import google.colab
        return True
    except ImportError:
        return False


def is_kaggle():
    # Is environment a Kaggle Notebook?
    try:
        assert os.environ.get('PWD') == '/kaggle/working'
        assert os.environ.get('KAGGLE_URL_BASE') == 'https://www.kaggle.com'
        return True
    except AssertionError:
        return False


def check_online():
    # Check internet connectivity
    import socket
    try:
        socket.create_connection(("1.1.1.1", 443), 5)  # check host accessibility
        return True
    except OSError:
        return False


def check_python(minimum='3.7.0'):
    # Check current python version vs. required python version
    check_version(platform.python_version(), minimum, name='Python ', hard=True)


def check_version(current='0.0.0', minimum='0.0.0', name='version ', pinned=False, hard=False, verbose=False):
    # Check version vs. required version
    current, minimum = (pkg.parse_version(x) for x in (current, minimum))
    result = (current == minimum) if pinned else (current >= minimum)  # bool
    s = f'{name}{minimum} required by YOLOv5, but {name}{current} is currently installed'  # string
    if hard:
        assert result, s  # assert min requirements met
    if verbose and not result:
        LOGGER.warning(s)
    return result


def try_except(func):
    # try-except function. Usage: @try_except decorator
    def handler(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            print(e)

    return handler


@try_except
def check_requirements(requirements=ROOT / 'requirements.txt', exclude=(), install=True):
    # Check installed dependencies meet requirements (pass *.txt file or list of packages)
    prefix = colorstr('red', 'bold', 'requirements:')
    check_python()  # check python version
    if isinstance(requirements, (str, Path)):  # requirements.txt file
        file = Path(requirements)
        assert file.exists(), f"{prefix} {file.resolve()} not found, check failed."
        with file.open() as f:
            requirements = [f'{x.name}{x.specifier}' for x in pkg.parse_requirements(f) if x.name not in exclude]
    else:  # list or tuple of packages
        requirements = [x for x in requirements if x not in exclude]

    n = 0  # number of packages updates
    for r in requirements:
        try:
            pkg.require(r)
        except Exception:  # DistributionNotFound or VersionConflict if requirements not met
            s = f"{prefix} {r} not found and is required by YOLOv5"
            if install:
                LOGGER.info(f"{s}, attempting auto-update...")
                try:
                    assert check_online(), f"'pip install {r}' skipped (offline)"
                    LOGGER.info(check_output(f"pip install '{r}'", shell=True).decode())
                    n += 1
                except Exception as e:
                    LOGGER.warning(f'{prefix} {e}')
            else:
                LOGGER.info(f'{s}. Please install and rerun your command.')

    if n:  # if packages updated
        source = file.resolve() if 'file' in locals() else requirements
        s = f"{prefix} {n} package{'s' * (n > 1)} updated per {source}\n" \
            f"{prefix} ⚠️ {colorstr('bold', 'Restart runtime or rerun command for updates to take effect')}\n"
        LOGGER.info(emojis(s))


def set_logging(name=None, verbose=VERBOSE):
    # Sets level and returns logger
    if is_kaggle():
        for h in logging.root.handlers:
            logging.root.removeHandler(h)  # remove all handlers associated with the root logger object
    rank = int(os.getenv('RANK', -1))  # rank in world for Multi-GPU trainings
    logging.basicConfig(format="%(message)s", level=logging.INFO if (verbose and rank in (-1, 0)) else logging.WARNING)
    return logging.getLogger(name)


# Globals
LOGGER = set_logging('yolov5')  # define globally (used in train.py, val.py, detect.py, etc.)
PREFIX = colorstr('Ultralytics: ')



# Global variables

# ENVIRONMENT = 'development'
ENVIRONMENT = 'production'

if ENVIRONMENT == 'production':
    HUB_API_ROOT = "https://api.ultralytics.com/"
    REPO_URL = "https://github.com/ultralytics/yolov5.git"
    REPO_BRANCH = "v6.1"  # master
else:
    HUB_API_ROOT = "http://127.0.0.1:5000/"
    # REPO_URL = "https://github.com/KalenMike/yolov5.git"
    # REPO_BRANCH = "feature/pip-package-clone"
    REPO_URL = "https://github.com/ultralytics/yolov5.git"
    REPO_BRANCH = "v6.1"  # master



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
PyTorch utils
"""

import datetime
import os
import platform
from pathlib import Path

import torch


def date_modified(path=__file__):
    # return human-readable file modification date, i.e. '2021-3-26'
    t = datetime.datetime.fromtimestamp(Path(path).stat().st_mtime)
    return f'{t.year}-{t.month}-{t.day}'


def select_device(device='', batch_size=None, newline=True, version='0.0.0'):
    # device = 'cpu' or '0' or '0,1,2,3'
    s = f'Ultralytics 🚀 {version or date_modified()} torch {torch.__version__} '  # string
    device = str(device).strip().lower().replace('cuda:', '')  # to string, 'cuda:0' to '0'
    cpu = device == 'cpu'
    if cpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # force torch.cuda.is_available() = False
    elif device:  # non-cpu device requested
        os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
        assert torch.cuda.is_available(), f'CUDA unavailable, invalid device {device} requested'  # check availability

    cuda = not cpu and torch.cuda.is_available()
    if cuda:
        devices = device.split(',') if device else '0'  # range(torch.cuda.device_count())  # i.e. 0,1,6,7
        n = len(devices)  # device count
        if n > 1 and batch_size:  # check batch_size is divisible by device_count
            assert batch_size % n == 0, f'batch-size {batch_size} not multiple of GPU count {n}'
        space = ' ' * (len(s) + 1)
        for i, d in enumerate(devices):
            p = torch.cuda.get_device_properties(i)
            s += f"{'' if i == 0 else space}CUDA:{d} ({p.name}, {p.total_memory / 1024 ** 2:.0f}MiB)\n"  # bytes to MB
    else:
        s += 'CPU\n'

    if not newline:
        s = s.rstrip()
    print(s.encode().decode('ascii', 'ignore') if platform.system() == 'Windows' else s)  # emoji-safe
    return torch.device('cuda:0' if cuda else 'cpu')



import re
from pathlib import Path
from urllib import request

from setuptools import find_packages, setup

# Settings
FILE = Path(__file__).resolve()
ROOT = FILE.parent  # YOLOv5 root directory

# The text of the README file
# README = (ROOT / "README.md").read_text(encoding="utf-8")
README = request.urlopen('https://raw.githubusercontent.com/ultralytics/hub/master/README.md').read().decode('utf-8')


def get_version():
    file = ROOT / 'src/ultralytics/__init__.py'
    return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', file.read_text(), re.M).group(1)


setup(
    name="ultralytics",  # name of pypi package
    version=get_version(),  # version of pypi package
    python_requires=">=3.7.0",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ultralytics/yolov5",
    author="Ultralytics",
    author_email='hello@ultralytics.com',
    package_dir={'': 'src'},  # Optional, use if source code is in a subdirectory under the project root, i.e. `src/`
    packages=find_packages('src'),  # Required
    include_package_data=True,
    install_requires=[
        'PyYAML>=5.3.1', 'requests', 'GitPython>=3.1.24', 'torch>=1.7.0', 'torchvision>=0.8.1', 'psutil', 'IPython'],
    extras_require={'tests': [
        'pytest',
        'pytest-cov',
        'coverage', ]},
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7", "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9", "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows"],
    keywords="machine-learning, deep-learning, ML, AI, PyTorch, object-detection, vision, YOLO, YOLOv3, YOLOv4, YOLOv5")



import requests

from .config import HUB_API_ROOT
from .hub_logger import HUBLogger
from .utils.general import PREFIX, emojis
from .yolov5_wrapper import YOLOv5Wrapper as YOLOv5


class Trainer:
    def __init__(self, model_id, auth):
        self.auth = auth
        self.model = self._get_model(model_id)
        if self.model is not None:
            self._connect_callbacks()

    def _get_model_by_id(self):
        # return a specific model
        return

    def _get_next_model(self):
        # return next model in queue
        return

    def _get_model(self, model_id):
        """
        Returns model from database by id
        """
        api_url = HUB_API_ROOT + "/model"
        payload = {"modelId": model_id}
        payload.update(self.auth.get_auth_string())

        try:
            r = requests.post(api_url, json=payload)
            res = r.json()

            if res["data"] is None:
                print(f"{PREFIX}ERROR: Unable to fetch model")
                return None  # Cannot train without model
            elif not res["data"]:
                print(
                    emojis(f"{PREFIX}No models to train. "
                           f"Create YOLOv5 🚀 models to train at https://hub.ultralytics.com"))
                return None  # Cannot train without model
            else:
                self.model_id = res["data"]["id"]  # Append id as it may be fetched from queue and unknown
                return res["data"]
        except requests.exceptions.ConnectionError:
            print(f'{PREFIX}ERROR: The HUB server is not online. Please try again later.')
            return None
            # sys.exit(141)  # TODO: keep this 141 sys exit error code?

    def _connect_callbacks(self):
        callback_handler = YOLOv5.new_callback_handler()
        hub_logger = HUBLogger(self.model_id, self.auth)
        callback_handler.register_action("on_pretrain_routine_end", "HUB", hub_logger.on_pretrain_routine_end)
        callback_handler.register_action("on_fit_epoch_end", "HUB", hub_logger.on_fit_epoch_end)
        callback_handler.register_action("on_model_save", "HUB", hub_logger.on_model_save)
        callback_handler.register_action("on_train_end", "HUB", hub_logger.on_train_end)
        self.callbacks = callback_handler

    def start(self):
        YOLOv5.train(self.callbacks, **self.model)
